//
// Syd: rock-solid unikernel
// src/pool.rs: Self growing / shrinking `ThreadPool` implementation
//
// Copyright (c) 2024 Ali Polatel <alip@chesswob.org>
// Based in part upon rusty_pool which is:
//     Copyright (c) Robin Friedli <robinfriedli@icloud.com>
//     SPDX-License-Identifier: Apache-2.0
//
// SPDX-License-Identifier: GPL-3.0-or-later

// Last sync with rusty_pool:
// Version 0.7.0
// Commit:d56805869ba3cbe47021d5660bbaf19ac5ec4bfb

use std::{
    mem::MaybeUninit,
    option::Option,
    os::fd::{BorrowedFd, RawFd},
    sync::{
        atomic::{AtomicU32, Ordering},
        Arc,
    },
    thread,
    time::Duration,
};

use libseccomp_sys::{
    seccomp_notif, seccomp_notif_resp, seccomp_notify_receive, seccomp_notify_respond,
};
use nix::{
    errno::Errno,
    poll::PollTimeout,
    sys::epoll::{Epoll, EpollCreateFlags, EpollEvent, EpollFlags},
    unistd::close,
};
use once_cell::sync::Lazy;
use parking_lot::RwLock;

use crate::{
    compat::epoll_ctl_safe,
    config::*,
    hook::{HandlerMap, UNotifyEventRequest},
    proc::proc_mmap,
    sandbox::Sandbox,
    ScmpNotifReq, SydArch, Sydcall,
};

#[allow(clippy::cast_possible_truncation)]
static NPROC: Lazy<u16> = Lazy::new(|| num_cpus::get() as u16);

const EOWNERDEAD: i32 = -nix::libc::EOWNERDEAD;

/// The absolute maximum number of workers. This corresponds to the
/// maximum value that can be stored within half the bits of u32, as two
/// counters (total workers and idle workers) are stored in one
/// AtomicU32.
pub(crate) const MAX_SIZE: u16 = u16::MAX;

/// Job is a combination of seccomp fd and seccomp request.
type Job = (RawFd, ScmpNotifReq);

/// Self growing / shrinking `ThreadPool` implementation based on
/// flume's multi-producer multi-consumer channels that enables awaiting
/// the result of a task and offers async support.
///
/// This `ThreadPool` has two different pool sizes; a core pool size
/// filled with threads that live for as long as the channel and a max
/// pool size which describes the maximum amount of worker threads that
/// may live at the same time. Those additional non-core threads have a
/// specific keep_alive time described when creating the `ThreadPool`
/// that defines how long such threads may be idle for without receiving
/// any work before giving up and terminating their work loop.
///
/// This `ThreadPool` does not spawn any threads until a task is
/// submitted to it. Then it will create a new thread for each task
/// until the core pool size is full. After that a new thread will only
/// be created upon an `execute()` call if the current pool is lower
/// than the max pool size and there are no idle threads.
///
/// Functions like `evaluate()` and `complete()` return a `JoinHandle`
/// that may be used to await the result of a submitted task or future.
/// JoinHandles may be sent to the thread pool to create a task that
/// blocks a worker thread until it receives the result of the other
/// task and then operates on the result. If the task panics the
/// `JoinHandle` receives a cancellation error. This is implemented
/// using a futures oneshot channel to communicate with the worker
/// thread.
///
/// When creating a new worker this `ThreadPool` tries to increment the
/// worker count using a compare-and-swap mechanism, if the increment
/// fails because the total worker count has been incremented to the
/// specified limit (the core_size when trying to create a core thread,
/// else MAX_SIZE) by another thread, the pool tries to create a
/// non-core worker instead (if previously trying to create a core
/// worker and no idle worker exists) or sends the task to the channel
/// instead. Panicking workers are always cloned and replaced.
///
/// Locks are only used for the join functions to lock the `Condvar`,
/// apart from that this `ThreadPool` implementation fully relies on
/// flume and atomic operations. This `ThreadPool` decides whether it is
/// currently idle (and should fast-return join attempts) by comparing
/// the total worker count to the idle worker count, which are two
/// values stored in one `AtomicU32` (both half the size of usize)
/// making sure that if both are updated they may be updated in a single
/// atomic operation.
///
/// The thread pool and its flume channel can be destroyed by using
/// the shutdown function, however that does not stop tasks that are
/// already running but will terminate the thread the next time it will
/// try to fetch work from the channel. The channel is only destroyed
/// once all clones of the `ThreadPool` have been shut down / dropped.
#[derive(Clone)]
pub(crate) struct ThreadPool {
    core_size: u16,
    keep_alive: u16,
    fd: RawFd,
    epoll: Arc<Epoll>,
    sandbox: Arc<RwLock<Sandbox>>,
    handlers: Arc<HandlerMap>,
    worker_data: Arc<WorkerData>,
}

impl ThreadPool {
    /// Construct a new `ThreadPool` with the specified core pool size,
    /// max pool size and keep_alive time for non-core threads. This
    /// function does not spawn any threads.
    ///
    /// `core_size` specifies the amount of threads to keep alive for as
    /// long as the `ThreadPool` exists and its channel remains
    /// connected.
    ///
    /// `keep_alive` specifies the duration in milliseconds for which to
    /// keep non-core pool worker threads alive while they do not
    /// receive any work.
    pub(crate) fn new(
        core_size: u16,
        keep_alive: u16,
        fd: RawFd,
        sandbox: Arc<RwLock<Sandbox>>,
        handlers: Arc<HandlerMap>,
    ) -> Result<Self, Errno> {
        // Create epoll instance.
        let epoll = Epoll::new(EpollCreateFlags::EPOLL_CLOEXEC)?;

        // Add seccomp fd to epoll (becomes readable when system call is interrupted).
        // We do not need the data field as we operate on a single fd.
        // Quoting: https://idea.popcount.org/2017-02-20-epoll-is-fundamentally-broken-12/
        // """
        // The best and the only scalable approach is to use recent
        // Kernel 4.5+ and use level-triggered events with
        // EPOLLEXCLUSIVE flag. This will ensure only one thread is
        // woken for an event, avoid "thundering herd" issue and scale
        // properly across multiple CPU's.
        // """
        let event = EpollEvent::new(EpollFlags::EPOLLIN | EpollFlags::EPOLLEXCLUSIVE, 0);
        // SAFETY: Borrow FD to make I/O safe API hippie.
        let epoll_fd = unsafe { BorrowedFd::borrow_raw(fd) };
        epoll_ctl_safe(&epoll, &epoll_fd, Some(event))?;

        Ok(Self {
            fd,
            sandbox,
            handlers,
            core_size,
            keep_alive,
            epoll: Arc::new(epoll),
            worker_data: Arc::new(WorkerData::default()),
        })
    }

    /// Boot the thread pool. This is the main entry point.
    pub(crate) fn boot(self) -> Result<(), std::io::Error> {
        // Return error if we cannot spawn the initial core thread.
        self.try_spawn()?;
        // Spawn the monitor thread to handle the rest.
        self.monitor()
    }

    /// Spawn a monitor thread that watches the worker pool idle count,
    /// and spawns new helper threads as necessary. This is done to
    /// ensure a sandbox process cannot DOS Syd by merely exhausting
    /// workers by e.g. opening the read end of a FIFO over and over
    /// again.
    pub(crate) fn monitor(self) -> Result<(), std::io::Error> {
        thread::Builder::new()
            .name("syd_mon".to_string())
            .stack_size(MON_STACK_SIZE)
            .spawn(move || {
                // Spawn the rest of the core threads.
                for _ in 1..*EMU_POOL_SIZE {
                    if matches!(self.try_spawn(), Ok(false) | Err(_)) {
                        break;
                    }
                }
                // Loop and spawn helper threads as necessary.
                let wait = Duration::from_millis(self.keep_alive.into());
                loop {
                    std::thread::sleep(wait);
                    let _ = self.try_spawn();
                }
            })
            .map(drop)
    }

    /// Try to create a new worker thread as needed.
    /// Returns Ok(true) if spawn succeeded, Ok(false) if no spawn was needed.
    pub(crate) fn try_spawn(&self) -> Result<bool, std::io::Error> {
        // Create a new worker either if the current worker count is
        // lower than the core pool size or if there are no idle threads
        // and the current worker count is lower than the max pool size.
        let worker_count_val = self.worker_data.0.load(Ordering::Relaxed);
        let (curr_worker_count, idle_worker_count) = WorkerData::split(worker_count_val);

        let keep_alive = if curr_worker_count < self.core_size {
            // Create a new core worker if current pool size is below
            // core size during the invocation of this function.
            None
        } else if curr_worker_count < MAX_SIZE && idle_worker_count == 0 {
            // Create a new helper worker if the current worker count is
            // below the MAX_SIZE and the pool has been observed to be
            // busy (no idle workers) during the invocation of this
            // function.
            Some(self.keep_alive)
        } else {
            // No need to spawn a new worker.
            return Ok(false);
        };

        // Try to spawn a new worker.
        Worker::new(
            self.fd,
            Arc::clone(&self.epoll),
            Arc::clone(&self.sandbox),
            Arc::clone(&self.handlers),
            keep_alive,
            Arc::clone(&self.worker_data),
        )
        .try_spawn()?;

        Ok(true)
    }
}

/// A helper struct to aid creating a new `ThreadPool` using default values where no value was
/// explicitly specified.
#[derive(Default)]
pub(crate) struct Builder {
    core_size: Option<u16>,
    keep_alive: Option<u16>,
}

impl Builder {
    /// Create a new `Builder`.
    pub(crate) fn new() -> Builder {
        Builder::default()
    }

    /// Specify the core pool size for the `ThreadPool`. The core pool
    /// size is the number of threads that stay alive for the entire
    /// lifetime of the `ThreadPool` or, to be more precise, its
    /// channel. These threads are spawned if a task is submitted to the
    /// `ThreadPool` and the current worker count is below the core pool
    /// size.
    pub(crate) fn core_size(mut self, size: u16) -> Builder {
        self.core_size = Some(size);
        self
    }

    /// Specify the duration in milliseconds for which additional
    /// threads outside the core pool remain alive while not receiving
    /// any work before giving up and terminating.
    pub(crate) fn keep_alive(mut self, keep_alive: u16) -> Builder {
        self.keep_alive = Some(keep_alive);
        self
    }

    /// Build the `ThreadPool` using the parameters previously supplied
    /// to this `Builder` using the number of CPUs as default core size
    /// if none provided, twice the core size as max size if none
    /// provided, 60 seconds keep_alive if none provided and the default
    /// naming (syd_{pool_number}) if none provided. This
    /// function calls
    /// [`ThreadPool::new`](struct.ThreadPool.html#method.new) or
    /// [`ThreadPool::new_named`](struct.ThreadPool.html#method.new_named)
    /// depending on whether a name was provided.
    pub(crate) fn build(
        self,
        fd: RawFd,
        sandbox: Arc<RwLock<Sandbox>>,
        handlers: Arc<HandlerMap>,
    ) -> Result<ThreadPool, Errno> {
        let core_size = self.core_size.unwrap_or(*EMU_POOL_SIZE);
        let keep_alive = self.keep_alive.unwrap_or(MON_KEEPALIVE);
        ThreadPool::new(core_size, keep_alive, fd, sandbox, handlers)
    }
}

#[derive(Clone)]
struct Worker {
    fd: RawFd,
    epoll: Arc<Epoll>,
    sandbox: Arc<RwLock<Sandbox>>,
    handlers: Arc<HandlerMap>,
    keep_alive: Option<u16>,
    worker_data: Arc<WorkerData>,
}

impl Worker {
    fn new(
        fd: RawFd,
        epoll: Arc<Epoll>,
        sandbox: Arc<RwLock<Sandbox>>,
        handlers: Arc<HandlerMap>,
        keep_alive: Option<u16>,
        worker_data: Arc<WorkerData>,
    ) -> Self {
        Worker {
            fd,
            epoll,
            sandbox,
            handlers,
            keep_alive,
            worker_data,
        }
    }

    fn try_spawn(self) -> Result<(), std::io::Error> {
        thread::Builder::new()
            .name("syd_emu".to_string())
            .stack_size(EMU_STACK_SIZE)
            .spawn(move || {
                // Create sentinel.
                let mut sentinel = Sentinel::new(&self);

                // Thread successfully started, increment total and idle
                // worker count, and set CPU affinity.
                let (worker_count, _) = self.worker_data.increment_both();
                // SAFETY: NPROC is never zero.
                #[allow(clippy::arithmetic_side_effects)]
                let core_id = usize::from(worker_count % *NPROC);
                core_affinity::set_for_current(core_affinity::CoreId { id: core_id });

                // Thread is a core thread if keep-alive is None.
                while let Ok(request) = self.receive() {
                    if let Some(request) = request {
                        // Mark current as no longer idle and process request.
                        self.worker_data.decrement_worker_idle();
                        self.process(&mut sentinel, request);
                    } // If None, child possibly died mid-way, retry with poll.
                } // If error, channel was broken because sender was
                  // disconnected.

                // Can decrement both at once as the thread only
                // gets here from an idle state (if waiting for work
                // and receiving an error).
                self.worker_data.decrement_both();
            })
            .map(drop)
    }

    fn receive(&self) -> Result<Option<ScmpNotifReq>, Errno> {
        // Wait for the request to become ready as necessary.
        // epoll_wait(2) will timeout and exit for non-core threads.
        self.poll()?;

        // Receive and return request.
        // Break if file descriptor was closed.
        // Ignore rest of the errors as we cannot handle them,
        // e.g: ENOENT: task is killed mid-way.
        match self.read() {
            Ok(request) => Ok(Some(request)),
            Err(Errno::EBADF) => Err(Errno::EBADF),
            Err(_) => Ok(None),
        }
    }

    fn read(&self) -> Result<ScmpNotifReq, Errno> {
        // Use libc::seccomp_notif rather than libseccomp_sys's.
        // The latter is opaque and required us to do a heap
        // allocation which we don't always want.
        let mut req: MaybeUninit<nix::libc::seccomp_notif> = MaybeUninit::zeroed();
        let ptr = req.as_mut_ptr() as *mut seccomp_notif;

        loop {
            // SAFETY: libseccomp's wrapper allocates each call.
            let ret = match unsafe { seccomp_notify_receive(self.fd, ptr) } {
                0 => {
                    // SAFETY: seccomp_notify_receive returned success.
                    // Request is populated and accessing it is safe.
                    ScmpNotifReq::from_sys(unsafe { req.assume_init() })
                }
                _ => Err(Errno::last()),
            };

            if ret != Err(Errno::EINTR) {
                return ret;
            }
        }
    }

    #[inline]
    fn process(&self, sentinel: &mut Sentinel, request: ScmpNotifReq) {
        sentinel.is_working = true;
        self.handle(request);
        sentinel.is_working = false;
        // can already mark as idle as this thread will continue the work loop.
        self.worker_data.increment_worker_idle();
    }

    #[allow(clippy::cognitive_complexity)]
    fn handle(&self, req: ScmpNotifReq) {
        let request = UNotifyEventRequest::new(req, self.fd, Arc::clone(&self.sandbox));
        let sydcall = Sydcall(req.data.syscall, req.data.arch);
        let handler = if let Some(handler) = self.handlers.get(&sydcall) {
            handler
        } else {
            unreachable!(
                "BUG: Missing hook for syscall {}!",
                i32::from(req.data.syscall)
            )
        };
        let resp = handler(&request);
        let mut resp = nix::libc::seccomp_notif_resp {
            id: resp.id,
            val: resp.val,
            error: resp.error,
            flags: resp.flags,
        };
        // libc's data type is not opaque unlike libseccomp's.
        let ptr: *mut seccomp_notif_resp = std::ptr::addr_of_mut!(resp) as *mut seccomp_notif_resp;

        loop {
            // SAFETY: libseccomp's version allocates needlessly...
            if unsafe { seccomp_notify_respond(self.fd, ptr) } == 0 || Errno::last() != Errno::EINTR
            {
                break;
            }
        }

        // EOWNERDEAD is a pseudo errno used by
        // the stat handler thread to close the
        // seccomp notify fd upon receiving the
        // "ghost" command.
        if resp.error == EOWNERDEAD {
            let _ = close(self.fd);

            let req = request.get_request();
            crate::info!("ctx": "main", "op": "ghost",
                "pid": req.pid, "sys": sydcall,
                "arch": SydArch(req.data.arch),
                "args": req.data.args,
                "src": proc_mmap(req.pid(), req.data.arch, req.data.instr_pointer));
        }
    }

    fn poll(&self) -> Result<(), Errno> {
        let timeout = if let Some(keep_alive) = self.keep_alive {
            PollTimeout::from(keep_alive)
        } else {
            PollTimeout::NONE
        };

        // Wait for an event and handle EINTR.
        let mut events = vec![EpollEvent::empty(); 1];
        loop {
            return match self.epoll.wait(&mut events, timeout) {
                Ok(0) if self.keep_alive.is_some() => Err(Errno::ETIMEDOUT),
                Ok(0) | Err(Errno::EINTR) => continue,
                Ok(_) if events[0].events().contains(EpollFlags::EPOLLIN) => Ok(()),
                Ok(_) => Err(Errno::EBADF), // POLLHUP!
                Err(errno) => Err(errno),
            };
        }
    }
}

/// Type that exists to manage worker exit on panic.
///
/// This type is constructed once per `Worker` and implements `Drop` to
/// handle proper worker exit in case the worker panics when executing
/// the current task or anywhere else in its work loop. If the
/// `Sentinel` is dropped at the end of the worker's work loop and the
/// current thread is panicking, handle worker exit the same way as if
/// the task completed normally (if the worker panicked while executing
/// a submitted task) then clone the worker and start it with an initial
/// task of `None`.
struct Sentinel<'s> {
    is_working: bool,
    worker_ref: &'s Worker,
}

impl Sentinel<'_> {
    fn new(worker_ref: &Worker) -> Sentinel<'_> {
        Sentinel {
            is_working: false,
            worker_ref,
        }
    }
}

impl Drop for Sentinel<'_> {
    fn drop(&mut self) {
        if thread::panicking() {
            if self.is_working {
                // Worker thread panicked.
                self.worker_ref.worker_data.decrement_worker_total();
            } else {
                // Idle thread panicked.
                self.worker_ref.worker_data.decrement_both();
            }
        }
    }
}

const WORKER_IDLE_MASK: u32 = 0x0000FFFF;
const INCREMENT_TOTAL: u32 = 0x00010000;
const INCREMENT_IDLE: u32 = 0x00000001;

/// 1. Struct containing data shared between workers.
/// 2. Struct that stores and handles an `AtomicU32` that stores the
///    total worker count in the higher half of bits and the idle worker
///    count in the lower half of bits. This allows to to increment /
///    decrement both counters in a single atomic operation.
#[derive(Default)]
struct WorkerData(AtomicU32);

impl WorkerData {
    #[allow(dead_code)]
    fn increment_both(&self) -> (u16, u16) {
        let old_val = self
            .0
            .fetch_add(INCREMENT_TOTAL | INCREMENT_IDLE, Ordering::Relaxed);
        Self::split(old_val)
    }

    fn decrement_both(&self) -> (u16, u16) {
        let old_val = self
            .0
            .fetch_sub(INCREMENT_TOTAL | INCREMENT_IDLE, Ordering::Relaxed);
        Self::split(old_val)
    }

    fn increment_worker_total(&self) -> u16 {
        let old_val = self.0.fetch_add(INCREMENT_TOTAL, Ordering::Relaxed);
        Self::total(old_val)
    }

    #[allow(dead_code)]
    fn decrement_worker_total(&self) -> u16 {
        let old_val = self.0.fetch_sub(INCREMENT_TOTAL, Ordering::Relaxed);
        Self::total(old_val)
    }

    fn increment_worker_idle(&self) -> u16 {
        let old_val = self.0.fetch_add(INCREMENT_IDLE, Ordering::Relaxed);
        Self::idle(old_val)
    }

    fn decrement_worker_idle(&self) -> u16 {
        let old_val = self.0.fetch_sub(INCREMENT_IDLE, Ordering::Relaxed);
        Self::idle(old_val)
    }

    fn get_total_count(&self) -> u16 {
        Self::total(self.0.load(Ordering::Relaxed))
    }

    fn get_idle_count(&self) -> u16 {
        Self::idle(self.0.load(Ordering::Relaxed))
    }

    #[inline]
    fn split(val: u32) -> (u16, u16) {
        let total_count = (val >> 16) as u16;
        let idle_count = (val & WORKER_IDLE_MASK) as u16;
        (total_count, idle_count)
    }

    #[inline]
    fn total(val: u32) -> u16 {
        (val >> 16) as u16
    }

    #[inline]
    fn idle(val: u32) -> u16 {
        (val & WORKER_IDLE_MASK) as u16
    }
}
